"""load word datasets to create word maps etc."""

import json
import os
import re
from typing import Any, Dict, Iterable, List, Set

from sign_language_translator.config.settings import Settings


class Vocab:
    """loads data from files for a specific language and specified sign_collections"""

    def __init__(
        self,
        language: str = r"$.",
        sign_collections: Iterable[str] = (r"$.",),
        data_root_dir: str = Settings.DATASET_ROOT_DIRECTORY,
        arg_is_regex: bool = True,
    ) -> None:
        # save arguments
        self.language = language
        self.sign_collections = sign_collections
        self.data_root_dir = data_root_dir
        self.arg_is_regex = arg_is_regex

        # source file paths
        self.label_to_words_path = os.path.join(
            self.data_root_dir,
            "sign_recordings",
            "collection_to_label_to_language_to_words.json",
        )
        self.constructed_words_path = os.path.join(
            self.data_root_dir,
            "sign_recordings",
            "organization_to_language_to_constructable_words.json",
        )
        self.preprocessing_path = os.path.join(
            self.data_root_dir,
            "text_preprocessing.json",
        )
        self.token_to_id_path = os.path.join(
            self.data_root_dir,
            "language_to_token_to_id.json",
        )

        # defaults
        self.word_sense_regex = r"\([^\(\)]*\)"  # everything in parenthesis
        self.word_to_labels: Dict[str, List[List[str]]] = {}
        self.token_to_id: Dict[str, int] = {}
        self.supported_words_with_word_sense: Set[str] = set()
        self.supported_words: Set[str] = set()
        self.ambiguous_to_unambiguous: Dict[str, List[str]] = {}
        self.ambiguous_to_unambiguous: Dict[str, List[str]] = {}

        # load jsons
        self.word_to_labels.update(
            self._make_word_to_labels(
                language=language,
                sign_collections=sign_collections,
                collection_to_label_to_language_to_words=self.__load_label_to_words(),
                organization_to_language_to_constructable_words=self.__load_constructable_words(),
                regex=arg_is_regex,
            )
        )
        self.token_to_id.update(self.__load_token_to_id(language, arg_is_regex))
        self.id_to_token: Dict[int, str] = {v: k for k, v in self.token_to_id.items()}

        # populate data from files
        self.supported_words_with_word_sense = set(self.word_to_labels)
        self.supported_words = {
            self.remove_word_sense(w) for w in self.supported_words_with_word_sense
        }
        self.ambiguous_to_unambiguous = self._make_disambiguation_map(
            self.supported_words_with_word_sense
        )
        self.labels: Set[str] = {
            label
            for label_sequences in self.word_to_labels.values()
            for label_sequence in label_sequences
            for label in label_sequence
        }

        preprocessing_map = self.__load_preprocessing(language)
        self.person_names: List[str] = preprocessing_map.get("person_names", [])
        self.words_to_numbers: Dict[str, int] = preprocessing_map.get(
            "words_to_numbers", {}
        )
        self.misspelled_to_correct: Dict[str, str] = preprocessing_map.get(
            "misspelled_to_correct", {}
        )
        self.number_suffix_to_zeros: Dict[str, str] = preprocessing_map.get(
            "number_suffixes_to_zeros", {}
        )
        self.joint_word_to_split_words: Dict[str, str] = preprocessing_map.get(
            "joint_word_to_split_words", {}
        )
        # TODO: improve coverage of key.isnumeric()
        self.numeric_keys: Set[str] = {
            key for key in self.word_to_labels if key.isnumeric()
        }

    def remove_word_sense(self, text: str) -> str:
        """Remove the word sense or disambiguation information from given text.

        Args:
            text (str): The text from which the word sense needs to be removed.

        Returns:
            str: The word without the word sense or disambiguation information.

        Example:
            word = "this is a spring(metal-coil). those are glasses(water-containers)."
            without_word_sense = remove_word_sense(word)
            print(without_word_sense)  # Output: "this is a spring. those are glasses."
        """

        without_word_sense = re.sub(self.word_sense_regex, "", text)
        return without_word_sense

    def __load_label_to_words(self):
        with open(
            self.label_to_words_path,
            "r",
            encoding="utf-8",
        ) as file_pointer:
            collection_to_label_to_language_to_words: Dict[
                str, Dict[str, Dict[str, List[str]]]
            ] = json.load(file_pointer)

        return collection_to_label_to_language_to_words

    def __load_constructable_words(self):
        with open(
            self.constructed_words_path,
            "r",
            encoding="utf-8",
        ) as file_pointer:
            organization_to_language_to_constructable_words: Dict[
                str, List[Dict[str, List[str]]]
            ] = json.load(file_pointer)

        return organization_to_language_to_constructable_words

    def __load_preprocessing(self, language: str, regex: bool = True) -> Dict[str, Any]:
        with open(
            self.preprocessing_path,
            "r",
            encoding="utf-8",
        ) as file_pointer:
            raw_data: Dict[str, Dict[str, Any]] = json.load(file_pointer)

        preprocessing_map: Dict[str, Any] = {
            key: lang_to_data.get(lang, self.__get_default_value(lang_to_data))
            for key, lang_to_data in raw_data.items()
            for lang in lang_to_data
            if self.__check_text_in_list(lang, [language], regex=regex)
        }

        return preprocessing_map

    def __load_token_to_id(self, language: str, regex: bool = True) -> Dict[str, int]:
        with open(
            self.token_to_id_path,
            "r",
            encoding="utf-8",
        ) as file_pointer:
            raw_data: Dict[str, Dict[str, int]] = json.load(file_pointer)

        token_to_id: Dict[str, int] = {
            key: val
            for lang, data in raw_data.items()
            for key, val in data.items()
            if self.__check_text_in_list(lang, [language], regex=regex)
        }

        return token_to_id

    def _make_word_to_labels(
        self,
        language: str,
        sign_collections: Iterable[str],
        collection_to_label_to_language_to_words: Dict[
            str, Dict[str, Dict[str, List[str]]]
        ],
        organization_to_language_to_constructable_words: Dict[
            str, List[Dict[str, List[str]]]
        ],
        regex: bool = True,
    ):
        """takes json word mapping datasets and creates a dict mapping text tokens to sign_labels

        Args:
            language (str | None, optional): language code used in json or a regex matching it whose data should be extracted. Defaults to None.
            sign_collections (Iterable[str] | None, optional): all sign collections to extract from json or regex matching them. Defaults to None.
            regex (bool, optional): treat the provided language code and sign_collections as regex when comparing against json keys. Defaults to True.

        Returns:
            Dict[str, List[List[str]]]: maps each text token to a list containing sequences of signs
        """
        word_to_labels: Dict[str, List[List[str]]] = {}

        # simple word map
        for (
            sign_coll,
            label_to_language_to_words,
        ) in collection_to_label_to_language_to_words.items():
            if not self.__check_text_in_list(sign_coll, sign_collections, regex=regex):
                continue
            for label, language_to_words in label_to_language_to_words.items():
                for lang, words in language_to_words.items():
                    if lang == "components" or not self.__check_text_in_list(
                        lang, [language], regex=regex
                    ):
                        continue
                    for word in words:
                        if word not in word_to_labels:
                            word_to_labels[word] = []

                        word_to_labels[word].append(
                            [f"{sign_coll}{Settings.FILENAME_SEPARATOR}{label}"]
                        )
                        if "components" in language_to_words:
                            # all components are from specified sign collections only
                            if all(
                                self.__check_text_in_list(
                                    comp.split(Settings.FILENAME_SEPARATOR)[0],
                                    sign_collections,
                                    regex=regex,
                                )
                                for comp in language_to_words["components"]
                            ):
                                word_to_labels[word].append(
                                    language_to_words["components"]
                                )

        # constructable words
        for (
            sign_coll,
            list_of_language_to_words,
        ) in organization_to_language_to_constructable_words.items():
            if not self.__check_text_in_list(sign_coll, sign_collections, regex=regex):
                continue
            for language_to_words in list_of_language_to_words:
                for lang, words in language_to_words.items():
                    if lang == "components" or not (
                        self.__check_text_in_list(lang, [language], regex=regex)
                        and "components" in language_to_words
                    ):
                        continue

                    for word in words:
                        if word not in word_to_labels:
                            word_to_labels[word] = []
                        word_to_labels[word].append(language_to_words["components"])

        # drop duplicates (when language = r".*", digits etc are repeated)
        word_to_labels = {
            word: list(list(seq) for seq in set(tuple(seq) for seq in label_seqs))
            for word, label_seqs in word_to_labels.items()
        }

        return word_to_labels

    def __check_text_in_list(
        self,
        text: str,
        list_of_texts: Iterable[str],
        regex: bool = True,
    ):
        if not (list_of_texts and all(list_of_texts)):
            return False

        return (
            any(re.match(pattern, text) for pattern in list_of_texts)
            if regex
            else text in list_of_texts
        )

    def __get_default_value(self, dictionary: Dict) -> Any:
        default_obj = None

        if dictionary:
            # get an empty instance of first value's class
            first_value = list(dictionary.values())[0]
            value_class = type(first_value)
            default_obj = value_class()

        return default_obj

    def _make_disambiguation_map(self, words: Iterable[str]) -> Dict[str, List[str]]:
        ambiguous_2_unambiguous = {}
        for word in words:
            without_word_sense = self.remove_word_sense(word)
            if without_word_sense != word:
                if without_word_sense not in ambiguous_2_unambiguous:
                    ambiguous_2_unambiguous[without_word_sense] = []
                ambiguous_2_unambiguous[without_word_sense].append(word)

        return ambiguous_2_unambiguous

    # load sign video/features


__all__ = [
    "Vocab",
]



"""transform landmarks
"""

from typing import List, Tuple, Union

import moviepy.editor as mpy
import numpy as np
import torch
from scipy.interpolate import interp1d as interpolate
from scipy.spatial.transform import Rotation as R
from skimage.transform import ProjectiveTransform, estimate_transform, warp
from tqdm.auto import tqdm

from ..utils.landmarks_info import LandmarksInfo


def _arg_type_normalizer(*args) -> Union[List[float], Tuple[float, float]]:
    """turn arguments into (float,) or ((float,float),)

    Returns:
        list: list of floats or list of tuples of 2 floats
    """

    scalars = []
    sequences = []

    return_sequence = False

    for arg in args:
        if hasattr(arg, "__len__") and len(arg) > 1:
            return_sequence = True
            sequences.append(list(arg))
        else:
            scalars.append(arg)
            sequences.append([arg, arg])

    return sequences if return_sequence else scalars


def zoom_landmarks(
    landmarks: Union[np.ndarray, torch.Tensor],
    x_factor: Union[float, Tuple[float, float]],
    y_factor: Union[float, Tuple[float, float]],
    z_factor: Union[float, Tuple[float, float]],
    add_random_steps=False,
    add_slow_spots=False,
) -> Union[np.ndarray, torch.Tensor]:
    """scale landmarks around their origin, separatly in each dimention

    Args:
        landmarks (Union[np.ndarray, torch.Tensor]): a 2D or 3D array of landmarks to be zoomed
        x_factor (Union[float, Tuple[float, float]]): scaling factor along x axis
        y_factor (Union[float, Tuple[float, float]]): scaling factor along y axis
        z_factor (Union[float, Tuple[float, float]]): scaling factor along z axis
        add_random_steps (bool, optional): True means add jitter to zooming so it is not smooth by sampling some values from uniform distribution. Defaults to False.
        add_slow_spots (bool, optional): True means add delays while zooming so it is not smooth by sample some values from a normal distribution concentrated around some point. Defaults to False.

    Raises:
        NotImplementedError: landmarks type other than numpy.ndarray or torch.Tensor are not handled.

    Returns:
        Union[np.ndarray, torch.Tensor]: scaled landmarks
    """
    # *np.random.normal(1,0.05,(3,2)).tolist() # for dynamic
    # *(np.random.rand(3)+0.5).tolist() # for static. right?

    x_factor, y_factor, z_factor = _arg_type_normalizer(x_factor, y_factor, z_factor)

    if not isinstance(x_factor, list):
        factor = [x_factor, y_factor, z_factor]
    else:
        n_frames = len(landmarks)
        # factor = np.linspace(
        #     [x_factor[0], y_factor[0], z_factor[0]],
        #     [x_factor[1], y_factor[1], z_factor[1]],
        #     n_frames,
        # )[:, np.newaxis, :]

        factor = np.concatenate(
            [
                _prepare_timesteps(
                    n_frames,
                    lower_limit=fac[0],
                    upper_limit=fac[1],
                    return_type=np.ndarray,
                    add_random_steps=add_random_steps,
                    add_slow_spots=add_slow_spots,
                )[:, np.newaxis, np.newaxis]
                for fac in [x_factor, y_factor, z_factor]
            ],
            axis=-1,
        )

    if isinstance(landmarks, torch.Tensor):
        factor = torch.Tensor(factor)

    elif isinstance(landmarks, np.ndarray):
        factor = np.array(factor)

    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    reshaper = LandmarksInfo.Reshaper()
    landmarks = reshaper.fold(landmarks)
    landmarks = landmarks * factor
    landmarks = reshaper.unfold(landmarks)

    return landmarks


def rotate_landmarks(
    landmarks: Union[np.ndarray, torch.Tensor],
    x_angle: Union[float, Tuple[float, float]],
    y_angle: Union[float, Tuple[float, float]],
    z_angle: Union[float, Tuple[float, float]],
    in_degrees: bool = True,
    add_random_steps=False,
    add_slow_spots=False,
) -> Union[np.ndarray, torch.Tensor]:
    """Rotate landmarks around their origin, with a separate angle for every dimention.

    Args:
        landmarks (Union[np.ndarray, torch.Tensor]): a 2D or 3D array of landmarks to be rotated
        x_angle (Union[float, Tuple[float, float]]): angle of rotation around x axis
        y_angle (Union[float, Tuple[float, float]]): angle of rotation around y axis
        z_angle (Union[float, Tuple[float, float]]): angle of rotation around z axis
        in_degrees (bool): True if angles are in degrees, False for radians.
        add_random_steps (bool, optional): True means add jitter to rotation so it is not smooth by sample some values from uniform distribution. Defaults to False.
        add_slow_spots (bool, optional): True means add delays in rotation so it is not smooth by  sample some values from a normal distribution concentrated around some point. Defaults to False.

    Raises:
        NotImplementedError: landmarks type other than numpy.ndarray or torch.Tensor are not handled.

    Returns:
        Union[np.ndarray, torch.Tensor]: rotated landmarks
    """
    # *np.random.normal(0,60,(3,2)).tolist(), in_degrees=True
    # *(np.random.rand(3)*360).tolist(), in_degrees=True # for static

    x_angle, y_angle, z_angle = _arg_type_normalizer(x_angle, y_angle, z_angle)

    if not isinstance(x_angle, list):
        rotation_matrix = R.from_euler(
            "xyz", [x_angle, y_angle, z_angle], degrees=in_degrees
        ).as_matrix()

    else:
        n_frames = len(landmarks)
        rotation_matrix = np.stack(
            [
                R.from_euler("xyz", [x, y, z], degrees=in_degrees).as_matrix()
                # for x, y, z in np.linspace(
                #     [x_angle[0], y_angle[0], z_angle[0]],
                #     [x_angle[1], y_angle[1], z_angle[1]],
                #     n_frames,
                # )
                for x, y, z in zip(
                    *[
                        _prepare_timesteps(
                            n_frames,
                            lower_limit=angle[0],
                            upper_limit=angle[1],
                            return_type=np.ndarray,
                            add_random_steps=add_random_steps,
                            add_slow_spots=add_slow_spots,
                        )
                        for angle in [x_angle, y_angle, z_angle]
                    ]
                )
            ]
        )

    if isinstance(landmarks, np.ndarray):
        matmul = np.matmul

    elif isinstance(landmarks, torch.Tensor):
        matmul = torch.matmul
        rotation_matrix = torch.from_numpy(rotation_matrix)

    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    reshaper = LandmarksInfo.Reshaper()
    landmarks = reshaper.fold(landmarks)
    landmarks = matmul(landmarks, rotation_matrix)
    landmarks = reshaper.unfold(landmarks)

    return landmarks


def noisy_landmarks(
    landmarks: Union[np.ndarray, torch.Tensor],
    vary_through_time: bool,
    nlandmarks_mean_std: List[Tuple[int, float, float]] = [
        [11, 0.003, 0.0005],  # face
        [4, 0.006, 0.0025],  # shoulders & elbows
        [2, 0.004, 0.0008],  # pose wrists
        [6, 0.0025, 0.0005],  # pose hands
        [10, 0.016, 0.0050],  # hips & legs
        [42, 0.002, 0.0004],  # hands
    ],  # [[33, 0.004, 0.0005], [42, 0.001, 0.0002]],
    visibility_noise_scale_mean: float = 0.015,
    visibility_noise_scale_std: float = 0.0025,
) -> Union[np.ndarray, torch.Tensor]:
    """Add gaussian noise of different scales to different landmarks.

    Args:
        landmarks (Union[np.ndarray, torch.Tensor]): a 2D or 3D array of landmarks
        vary_through_time (bool): True means the landmarks would get different noise value in diffent timesteps (landmarks vibrate as a result). False means same noise values added through all time steps.
        nlandmarks_mean_std (List[Tuple[int, float, float]], optional): each row tells how many landmarks should get noise of a scale determined by another gaussian distribution of that given mean and std. Defaults to [ [11, 0.003, 0.0005],  # face [4, 0.006, 0.0025],  # shoulders & elbows [2, 0.004, 0.0008],  # pose wrists [6, 0.0025, 0.0005],  # pose hands [10, 0.016, 0.0050],  # hips & legs [42, 0.002, 0.0004],  # hands ].
        visibility_noise_scale_mean (float, optional): the mean used by a gaussian distribution to determine the scale of noise added to the visibility feature. Defaults to 0.025.
        visibility_noise_scale_std (float, optional): the standard deviation used by a gaussian distribution to determine the scale of noise added to the visibility feature. Defaults to 0.010.

    Raises:
        NotImplementedError: landmarks type other than numpy.ndarray or torch.Tensor are not handled.

    Returns:
        Union[np.ndarray, torch.Tensor]: landmarks with noise added (which can be constant through time or changing)
    """

    if isinstance(landmarks, np.ndarray):
        normal = np.random.normal
        to_array = np.array

    elif isinstance(landmarks, torch.Tensor):
        normal = torch.normal
        to_array = torch.Tensor

    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    no_detections_mask = landmarks == 0
    reshaper = LandmarksInfo.Reshaper()
    landmarks = reshaper.fold(landmarks)
    landmark_dim = -2
    time_dim = -3

    # determine noise scales
    nlandmarks, means, stds = np.array(nlandmarks_mean_std).T
    noise_scales = np.random.normal(0, 1, len(nlandmarks))
    noise_scales *= stds
    noise_scales += means
    noise_scales = np.abs(noise_scales).repeat(nlandmarks.tolist()).reshape((-1, 1))
    noise_scales = to_array(noise_scales)

    # determine noise shape
    noise_size = list(landmarks.shape)
    if not vary_through_time:
        noise_size[time_dim] = 1

    # generate noise
    noise = normal(0, 1, noise_size)
    noise *= noise_scales

    assert (
        noise.shape[landmark_dim] == landmarks.shape[landmark_dim]
    ), "fix nlandmarks_mean_std to cover all landmarks"

    if reshaper.landmarks_visibility_backup is not None:
        visibility_noise_scale = abs(
            normal(visibility_noise_scale_mean, visibility_noise_scale_std, (1,))
        )
        visibility_shape = list(reshaper.landmarks_visibility_backup.shape)
        if not vary_through_time:
            visibility_shape[time_dim] = 1

        visibility_noise = normal(0, visibility_noise_scale, visibility_shape)
        reshaper.landmarks_visibility_backup = (
            reshaper.landmarks_visibility_backup + visibility_noise
        ).clip(0, 1)

    landmarks = landmarks + noise
    landmarks = reshaper.unfold(landmarks)
    landmarks[no_detections_mask] = 0

    return landmarks


def _prepare_timesteps(
    n_steps: int,
    lower_limit: float = 0,
    upper_limit: float = 1,
    return_type=np.ndarray,
    add_random_steps: bool = False,
    n_random_steps_frac: float = None,
    min_n_linear_steps: int = 20,
    min_n_linear_steps_frac: float = 0.25,
    add_slow_spots: bool = False,
    n_slow_spots: int = None,
    n_slow_spots_ratio: float = 0.02,
    slow_spots_steps_frac: float = 0.25,
) -> Union[np.ndarray, torch.Tensor]:
    """A mess of a function. Generates numbers in a given range which are evenly or randomly distributed or concentrated around some point(s).

    Args:
        n_steps (int): how many values to generate in the given range.
        lower_limit (float, optional): the lower bound of the range. Defaults to 0.
        upper_limit (float, optional): the upper bound of the range. Defaults to 1.
        return_type (type, optional): the type of object to return. Only np.ndarray and torch.Tensor are supported. Defaults to np.ndarray.
        add_random_steps (bool, optional): True means sample some values from uniform distribution. Defaults to False.
        n_random_steps_frac (float, optional): what percentage of total steps should be sampled from uniform distribution. If None, the ratio is randomly selected. Defaults to None.
        min_n_linear_steps (int, optional): minimum number of steps that should be evenly spaced and span the entire range. Defaults to 20.
        min_n_linear_steps_frac (float, optional): percentage of total steps at least which should be evenly spaced and span the entire range. Its an alternate to the above and maximum of both is used. Defaults to 0.25.
        add_slow_spots (bool, optional): True means sample some values from a normal distribution concentrated around some point. Defaults to False.
        n_slow_spots (int, optional): number of such concentrated points to use. If None, its selected randomly. Defaults to None.
        n_slow_spots_frac (float, optional): ratio of number of spots to total steps. Defaults to 0.02.
        slow_spots_steps_frac (float, optional): percentage of total steps to be used to make such spots. Defaults to 0.2.

    Raises:
        Exception: for some reason, no kind of values were generated.
        NotImplementedError: landmarks type other than numpy.ndarray or torch.Tensor are not handled.

    Returns:
        np.ndarray | torch.Tensor: an array of n sorted values between the given range.
    """

    steps = []
    remaining_steps = n_steps
    min_n_linear_steps = max(min_n_linear_steps, min_n_linear_steps_frac * n_steps)
    min_n_linear_steps = min(min_n_linear_steps, n_steps)

    if add_slow_spots and (remaining_steps > min_n_linear_steps):
        if n_slow_spots is None:
            n_slow_spots = round(
                n_steps * n_slow_spots_ratio + np.random.rand(1).item() * 0.8 - 0.4
            )

        if n_slow_spots > 0:
            total_slow_spots_steps = slow_spots_steps_frac * n_steps
            slow_spot_sizes = np.round(
                total_slow_spots_steps
                / n_slow_spots
                * (np.random.rand(n_slow_spots) * 0.8 + 0.6)
            )

            slow_spot_locations = (
                np.random.rand(n_slow_spots) * (upper_limit - lower_limit) + lower_limit
            ).repeat(slow_spot_sizes.tolist())

            slow_spot_scales = (
                slow_spot_sizes / n_steps / 4.5 * (upper_limit - lower_limit)
            ).repeat(slow_spot_sizes.tolist())

            total_slow_spots_steps = int(
                min([remaining_steps - min_n_linear_steps, np.sum(slow_spot_sizes)])
            )

            slow_steps = np.random.normal(0, 1, total_slow_spots_steps)
            slow_steps *= slow_spot_scales[:total_slow_spots_steps]
            slow_steps += slow_spot_locations[:total_slow_spots_steps]
            slow_steps = np.abs(slow_steps)

            steps.append(slow_steps)
            remaining_steps -= len(slow_steps)

    if add_random_steps:
        if n_random_steps_frac is None:
            n_random_steps_frac = np.random.rand(1) * 0.85 + 0.15

        n_random_steps = int(
            (remaining_steps - min_n_linear_steps) * n_random_steps_frac
        )

        steps.append(
            np.random.rand(n_random_steps) * (upper_limit - lower_limit) + lower_limit
        )
        remaining_steps -= n_random_steps

    steps.append(np.linspace(lower_limit, upper_limit, remaining_steps))

    if len(steps) == 1:
        steps = steps[0]
    elif len(steps) > 1:
        steps = np.concatenate(steps)
    else:
        raise Exception("empty timesteps list")

    assert len(steps) == n_steps

    if return_type == np.ndarray:
        steps = np.sort(steps)
    elif return_type == torch.Tensor:
        steps = torch.Tensor(np.sort(steps))
    else:
        raise NotImplementedError("use np.ndarray or torch.Tensor as return_type")

    return steps


def _interpolate_multi_frame_landmarks(
    multi_frame_landmarks: Union[np.ndarray, torch.Tensor],
    old_timesteps: Union[np.ndarray, torch.Tensor],
    new_timesteps: Union[np.ndarray, torch.Tensor],
    time_dimention: int = 0,
    bad_detection_threshold: float = 0.1,
) -> Union[np.ndarray, torch.Tensor]:
    """Interpolate new points within the given data across a given dimention. Should create new frames across time dimention at certain timestamps.

    Args:
        multi_frame_landmarks (Union[np.ndarray, torch.Tensor]): a 3D array of landmarks to be interpolated
        old_timesteps (Union[np.ndarray, torch.Tensor]): independent (time) axis values that correspond to each subarray in data.
        new_timesteps (Union[np.ndarray, torch.Tensor]): independent (time) axis values for which subarrays should be created.
        time_dimention (int, optional): the dimention number of the landmarks array across which interpolation should take place. Defaults to 0.
        bad_detection_threshold (float, optional): Used to convert interpolated mask into anther mask. values in interpolated mask greater than this are converted to True and corresponding landmarks put equal to 0. Defaults to 0.1.

    Raises:
        NotImplementedError: landmarks type other than numpy.ndarray or torch.Tensor are not handled.

    Returns:
        Union[np.ndarray, torch.Tensor]: Interpolated Landmarks
    """

    assert len(old_timesteps) == multi_frame_landmarks.shape[time_dimention]

    no_detection_mask = multi_frame_landmarks == 0
    if no_detection_mask.shape[-1] in [
        LandmarksInfo.N_ALL_FEATURES,
        LandmarksInfo.N_POSE_FEATURES,
    ]:
        no_detection_mask[..., LandmarksInfo.VISIBILITY_INDEXES] = True

    f_landmarks = interpolate(old_timesteps, multi_frame_landmarks, axis=time_dimention)
    f_mask = interpolate(old_timesteps, no_detection_mask, axis=time_dimention)

    new_landmarks = f_landmarks(new_timesteps)
    new_mask = f_mask(new_timesteps) >= bad_detection_threshold

    new_landmarks[new_mask] = 0

    if isinstance(multi_frame_landmarks, np.ndarray):
        return np.array(new_landmarks)
    elif isinstance(multi_frame_landmarks, torch.Tensor):
        return torch.Tensor(new_landmarks)
    else:
        raise NotImplementedError("Use np.ndarray or torch.Tensor")


def change_duration(
    multi_frame_landmarks: Union[np.ndarray, torch.Tensor],
    factor: float = None,
    n_new_timesteps: int = None,
    time_dimention: int = 0,
    bad_detection_threshold: float = 0.1,
    use_random_speed: bool = False,
    random_factor_range: Tuple[float, float] = (0.5, 1.5),
    add_random_steps: bool = False,
    n_random_steps_frac: float = None,
    min_n_linear_steps: int = 20,
    min_n_linear_steps_frac: float = 0.25,
    add_slow_spots: bool = False,
    n_slow_spots: int = None,
    n_slow_spots_frac: float = 0.02,
    slow_spots_steps_frac: float = 0.2,
    return_new_timesteps: bool = False,
) -> Union[
    np.ndarray,
    torch.Tensor,
    Tuple[np.ndarray, np.ndarray],
    Tuple[torch.Tensor, torch.Tensor],
]:
    """increase or decrease the number of subarrays (frames) across a dimention (time) by interpolating new data within the given data.

    Args:
        multi_frame_landmarks (Union[np.ndarray, torch.Tensor]): a 3D array of landmarks to be interpolated
        factor (float, optional): multiplier of old duration to get new duration. Defaults to None.
        n_new_timesteps (int, optional): the new duration. Alternate to factor, overwrites that. Defaults to None.
        time_dimention (int, optional): the dimention number of the landmarks array across which interpolation should take place. Defaults to 0.
        bad_detection_threshold (float, optional): Used to convert interpolated mask into anther mask. values in interpolated mask greater than this are converted to True and corresponding landmarks put equal to 0. Defaults to 0.1.
        use_random_speed (bool, optional): determine a factor by random. only works if Nones are provided as factor/n_new_timesteps. Defaults to False.
        random_factor_range (Tuple[float, float], optional): uniformly sample a factor value from this range. Defaults to (0.5, 1.5).
        add_random_steps (bool, optional): True means sample some values from uniform distribution. Defaults to False.
        n_random_steps_frac (float, optional): what percentage of total steps should be sampled from uniform distribution. If None, the ratio is randomly selected. Defaults to None.
        min_n_linear_steps (int, optional): minimum number of steps that should be evenly spaced and span the entire range. Defaults to 20.
        min_n_linear_steps_frac (float, optional): percentage of total steps at least which should be evenly spaced and span the entire range. Its an alternate to the above and maximum of both is used. Defaults to 0.25.
        add_slow_spots (bool, optional): True means sample some values from a normal distribution concentrated around some point. Defaults to False.
        n_slow_spots (int, optional): number of such concentrated points to use. If None, its selected randomly. Defaults to None.
        n_slow_spots_frac (float, optional): ratio of number of spots to total steps. Defaults to 0.02.
        slow_spots_steps_frac (float, optional): percentage of total steps to be used to make such spots. Defaults to 0.2.
        return_new_timesteps (bool, optional): True means the new timesteps array to which the interpolation corresponds is also returned. Defaults to False.

    Raises:
        ValueError: speed factor, n_new_timesteps, and use_random_speed are all None/False

    Returns:
        np.ndarray | torch.Tensor | Tuple[np.ndarray, np.ndarray] | Tuple[torch.Tensor, torch.Tensor]: The interpolated landmarks and/or the timesteps they correspond to.
    """

    n_old_timesteps = multi_frame_landmarks.shape[time_dimention]

    old_timesteps = _prepare_timesteps(
        n_old_timesteps,
        lower_limit=0,
        upper_limit=1,
        return_type=type(multi_frame_landmarks),
        add_random_steps=False,
        add_slow_spots=False,
    )

    if n_new_timesteps is None:
        if factor is None:
            if use_random_speed:
                factor = (
                    np.random.rand(1)
                    * (random_factor_range[1] - random_factor_range[0])
                    + random_factor_range[0]
                )
            else:
                raise ValueError("provide a speed factor or n_new_timesteps")
        n_new_timesteps = int(n_old_timesteps * factor)

    new_timesteps = _prepare_timesteps(
        n_new_timesteps,
        lower_limit=0,
        upper_limit=1,
        return_type=type(multi_frame_landmarks),
        add_random_steps=add_random_steps,
        n_random_steps_frac=n_random_steps_frac,
        min_n_linear_steps=min_n_linear_steps,
        min_n_linear_steps_frac=min_n_linear_steps_frac,
        add_slow_spots=add_slow_spots,
        n_slow_spots=n_slow_spots,
        n_slow_spots_ratio=n_slow_spots_frac,
        slow_spots_steps_frac=slow_spots_steps_frac,
    )

    new_landmarks = _interpolate_multi_frame_landmarks(
        multi_frame_landmarks,
        old_timesteps=old_timesteps,
        new_timesteps=new_timesteps,
        time_dimention=time_dimention,
        bad_detection_threshold=bad_detection_threshold,
    )

    return new_landmarks if not return_new_timesteps else (new_landmarks, new_timesteps)


def stabilize_landmarks(
    landmarks_list: List[Union[np.ndarray, torch.Tensor]],
    landmarks_to_use: List[int] = [11, 12, 23, 24],
    coordinates_to_use: List[int] = [0, 1, 2],
    transformation_type: str = "similarity",
    target: Union[List[List[float]], np.ndarray, torch.Tensor] = None,
    infer_target: bool = True,
    return_transformation_matrixes_only: bool = False,
) -> List[Union[np.ndarray, torch.Tensor]]:
    """Apply transformations to a list of multi_frame_landmarks such that they all land on the average location.

    Args:
        landmarks_list (List[Union[np.ndarray, torch.Tensor]]): list of multi_frame_landmarks which are 2D or 3D sequences of body landmarks
        landmarks_to_use (List[int], optional): the index of landmarks that should be used stabilization. Defaults to (torso) [11, 12, 23, 24].
        coordinates_to_use (List[int], optional): index of coordinates (x, y, z) that should be used in stabilization. choose depending on the type of landmarks. for example use x,y for Image landmarks and use x,y,z for 3D world landmarks. Defaults to [0, 1, 2].
        transformation_type (str, optional): parameter for skimage.estimate_transform which determines the allowed degrees of freedom. Choose from ['euclidean', 'similarity', 'affine', 'piecewise-affine', 'projective', 'polynomial']. Defaults to "similarity" which includes shift, scale, rotate.
        target (Union[List[List[float]], np.ndarray, torch.Tensor], optional): any desired destination for the transformed landmarks of shape (n_landmarks, n_coordinates). useful for rectification. Defaults to None.
        infer_target (bool, optional): True means that an average of the provided landmark numbers in the landmarks_list will be calculated and that will be the destination of the transformation. Needs the target argument to be None to work. Defaults to True.
        return_transformation_matrixes_only (bool, optional): True means that the estimated transformations for each multi_frame_landmarks in the landmarks_list will be calculated instead of the stabilized landmarks. Defaults to False.

    Raises:
        NotImplementedError: use np.ndarray or torch.Tensor landmarks
        ValueError: Either provide a target matrix or set infer_target to True.

    Returns:
        List[np.ndarray | torch.Tensor]: list of stabilized landmarks or the stabilzation matrices
    """

    if len(landmarks_list) == 0:
        return landmarks_list

    reshapers = [LandmarksInfo.Reshaper() for _ in range(len(landmarks_list))]
    landmarks_list = [
        reshaper.fold(landmarks)
        for landmarks, reshaper in zip(landmarks_list, reshapers)
    ]

    # reshape, concatenate, average, array, dot
    if isinstance(landmarks_list[0], np.ndarray):
        concatenate = np.concatenate
        average = lambda x, axis=0: np.average(x, axis=axis, weights=x != 0)
        to_array = np.array
        ones = np.ones

    elif isinstance(landmarks_list[0], torch.Tensor):
        concatenate = torch.concatenate
        average = lambda x, axis=0: torch.mean(x, dim=axis)
        to_array = torch.Tensor
        ones = torch.ones

    else:
        raise NotImplementedError("use numpy.ndarray or torch.Tensor")

    landmarks_to_use = to_array(landmarks_to_use).reshape((-1, 1)).tolist()

    if target is None:
        if infer_target:
            stacked = concatenate(landmarks_list, axis=0)[
                ..., landmarks_to_use, coordinates_to_use
            ]
            target = average(stacked, axis=0)
        else:
            ValueError("Either provide a target matrix or set infer_target to True.")

    elif isinstance(target, list):
        target = to_array(target)
        if target.shape != (
            len(landmarks_to_use),
            len(coordinates_to_use),
        ):
            raise ValueError(
                "target shape does not match specified landmarks or coordinates"
            )

    transformations_list = []

    for landmarks in landmarks_list:
        landmarks_ = landmarks[:, landmarks_to_use, coordinates_to_use]
        source = average(landmarks_, axis=0)

        tform = estimate_transform(
            transformation_type, np.array(source), np.array(target)
        )

        if not return_transformation_matrixes_only:
            M = to_array(tform.params)
            homogeneous = concatenate(
                [
                    landmarks[..., coordinates_to_use],
                    ones(landmarks.shape[:-1] + (1,)),
                ],
                axis=-1,
            )
            # homogeneous = homogeneous @ M.T
            # landmarks[..., coordinates_to_use] = homogeneous[..., :-1] / homogeneous[..., -1:]
            # dropping the bottom row of matrix because its just [0 0 0 1] and requires unnecesary divide operation later
            landmarks[..., coordinates_to_use] = homogeneous @ M[:-1].T

        else:
            transformations_list.append(to_array(tform.params))

    landmarks_list = [
        reshaper.unfold(landmarks)
        for landmarks, reshaper in zip(landmarks_list, reshapers)
    ]

    return (
        landmarks_list
        if not return_transformation_matrixes_only
        else transformations_list
    )


def transform_clip(
    clip: mpy.VideoFileClip,
    transformation_matrix: np.ndarray,
    border_mode: str = "reflect",
    show_progress: bool = True,
) -> mpy.VideoFileClip:
    """Apply a transformation matrix to a moviepy video clip.

    Args:
        clip (mpy.VideoFileClip): moviepy clips that needs to be transformed.
        transformation_matrix (np.ndarray): the transformation matrix according to which the frames should be transformed
        border_mode (str, optional): what pixels should be used to fill the borders when a frame is warped and shrinks from a side. valid values are ["reflect", "constant", "edge", "symmetric", "wrap"]. Defaults to "reflect".
        show_progress (bool, optional): display a progress bar showing how many clips are processed. Defaults to True.

    Returns:
        mpy.VideoFileClip: A transformed video clip
    """

    frames = clip.iter_frames()
    if show_progress:
        frames = tqdm(
            frames,
            desc="Processed frames",
            total=int(clip.fps * clip.duration),
            leave=False,
        )

    inverse_map = ProjectiveTransform(transformation_matrix).inverse
    clip = mpy.ImageSequenceClip(
        [
            # cv2.warpPerspective(frame, transformation_matrix, clip.size)
            warp(frame, inverse_map, mode=border_mode, preserve_range=True)
            for frame in frames
        ],
        fps=clip.fps,
    )
    clip = clip.set_audio(clip.audio)

    return clip


def stabilize_clips(
    clips: List[mpy.VideoFileClip],
    landmarks_list: List[Union[np.ndarray, torch.Tensor]],
    landmarks_to_use: List[int] = [11, 12, 23, 24],
    coordinates_to_use: List[int] = [0, 1],
    transformation_type: str = "similarity",
    target: Union[List[List[float]], np.ndarray, torch.Tensor] = None,
    infer_target: bool = True,
    scale_landmarks: bool = True,
    border_mode: str = "reflect",
    show_progress: bool = True,
) -> List[mpy.VideoFileClip]:
    """Move all the clips in a way that certain body regions are always on the same location.

    Args:
        clips (List[mpy.VideoFileClip]): list of moviepy clips that need to be stabilized
        landmarks_list (List[Union[np.ndarray, torch.Tensor]]): list of reference multi_frame_landmarks which were extracted from the clips
        landmarks_to_use (List[int], optional): the index of landmarks that should be used stabilization. Defaults to (torso) [11, 12, 23, 24].
        coordinates_to_use (List[int], optional): index of coordinates (x, y, z) that should be used in stabilization. choose depending on the type of landmarks. for example use x,y for Image landmarks and use x,y,z for 3D world landmarks. Defaults to [0, 1, 2].
        transformation_type (str, optional): parameter for skimage.estimate_transform which determines the allowed degrees of freedom. Choose from ['euclidean', 'similarity', 'affine', 'piecewise-affine', 'projective', 'polynomial']. Defaults to "similarity" which includes shift, scale, rotate.
        target (Union[List[List[float]], np.ndarray, torch.Tensor], optional): any desired destination for the transformed landmarks of shape (n_landmarks, n_coordinates). Defaults to None.
        infer_target (bool, optional): True means that an average of the provided landmark numbers in the landmarks_list will be calculated and that will be the destination of the transformation. Needs the target argument to be None to work. Defaults to True.
        scale_landmarks (bool, optional): if the landmarks are normalized by clip height and width, True means rescale them to original pixel size. Defaults to True.
        border_mode (str, optional): what pixels should be used to fill the borders when a frame is warped and shrinks from a side. valid values are ["reflect", "constant", "edge", "symmetric", "wrap"]. Defaults to "reflect"
        show_progress (bool, optional): display a progress bar showing how many clips are processed. Defaults to True.

    Returns:
        List[mpy.VideoFileClip]: list of stabilized moviepy clips
    """

    if scale_landmarks:
        reshaper = LandmarksInfo.Reshaper()
        _landmarks_list = []
        for clip, landmarks in zip(clips, landmarks_list):
            landmarks = np.array(landmarks)
            landmarks = reshaper.fold(landmarks)
            _landmarks_list.append(reshaper.unfold(landmarks * (list(clip.size) + [1])))
        landmarks_list = _landmarks_list

    transformations = stabilize_landmarks(
        landmarks_list,
        landmarks_to_use=landmarks_to_use,
        coordinates_to_use=coordinates_to_use,
        transformation_type=transformation_type,
        target=target,
        infer_target=infer_target,
        return_transformation_matrixes_only=True,
    )

    clip_matrix_pairs = zip(clips, transformations)
    if show_progress:
        clip_matrix_pairs = tqdm(
            clip_matrix_pairs,
            desc="Clips stabilized",
            total=len(clips),
            leave=False,
        )

    clips = [
        transform_clip(clip, matrix, border_mode=border_mode, show_progress=False)
        for clip, matrix in clip_matrix_pairs
    ]

    return clips



"""text scraping - search usable sentences, phrases or ngrams from corpora.
"""

import json
import re
import threading
import urllib.parse
from time import sleep
from typing import Dict, List, Set, Union

import requests
from bs4 import BeautifulSoup
from IPython.display import clear_output
from tqdm.auto import tqdm

captcha_waiting = False


def detect_captcha(page: str) -> bool:
    soup = BeautifulSoup(page, "html.parser")
    divs = soup.find_all("div", {"class": "g-recaptcha"})

    captcha_solving_needed = len(divs) > 0

    return captcha_solving_needed


def multi_threaded_requests(
    target_function, args_list, time_delay=0.02, max_n_threads=None
):
    threads = []
    for args in (bar := tqdm(args_list, desc="sending requests")):
        thread = threading.Thread(target=target_function, args=args)
        thread.start()
        threads.append(thread)
        sleep(time_delay)

        if max_n_threads:
            if len(threads) >= max_n_threads:
                for i, thread in enumerate(threads):
                    bar.set_description(f"waiting for threads {i/len(threads):.0%}")
                    thread.join()
                threads = []
                bar.set_description("sending requests")

    for thread in tqdm(threads, desc="waiting for threads"):
        thread.join()


def request_google(query: str, search_exact: bool = True) -> str:
    if search_exact:
        query = f'"{query}"'

    query = urllib.parse.quote(query)
    url = f"https://www.google.com/search?q={query}"

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
    }
    page = requests.get(url, headers=headers).text

    return page


def extract_google_results_count(page: str) -> int:
    n_results = 0

    soup = BeautifulSoup(page, "html.parser")
    divs = soup.find_all("div", {"id": "result-stats"})

    if divs:
        text = divs[0].next_element.text.replace(",", "")
        numbers = re.findall(r"\d+", text)
        if numbers:
            n_results = int(numbers[0])

    return n_results


def get_google_results_count(query: str, search_exact: bool = True) -> int:
    while True:
        response = request_google(query, search_exact=search_exact)

        if detect_captcha(response):
            print("!!!!!!!CAPTCHA!!!!!!!", query)
            sleep(2)
        else:
            break
    n_results = extract_google_results_count(response)

    return n_results


def request_glosbe(
    query: str, source_language: str, target_language: str, page_no: int
) -> str:

    url = f"https://glosbe.com/{source_language}/{target_language}/{urllib.parse.quote(query)}/fragment/tmem?page={page_no}&mode=MUST&stem=true"
    page = requests.get(url).text

    return page


def extract_glosbe_texts(page: str, element="span", css_class="dense") -> List[str]:
    soup = BeautifulSoup(page, "html.parser")
    cards = soup.find_all(element, {"class": css_class})

    texts = list({card.text for card in cards})

    return texts


def get_glosbe_texts(
    query: str,
    source_language: str,
    target_language: str,
    page_no: int,
    element="span",
    css_class="dense",
):
    global captcha_waiting
    while True:
        if captcha_waiting:
            sleep(10)

        response = request_glosbe(query, source_language, target_language, page_no)

        if detect_captcha(response):
            captcha_waiting = True
            print("!!!!!!!CAPTCHA!!!!!!!", query)
            sleep(10)
        else:
            clear_output()
            captcha_waiting = False
            break
    texts = extract_glosbe_texts(response, element=element, css_class=css_class)

    return texts


def scraping_function_for_glosbe(
    word: str,
    source_language: str,
    target_language: str,
    results: Dict[str, Union[Dict[str, int], Set[str]]],
):
    if results["next_page_number"][word] > 0:
        texts = get_glosbe_texts(
            word, source_language, target_language, results["next_page_number"][word]
        )
        with threading.Lock():
            if len(texts) > 0:
                results["next_page_number"][word] += 1
                results["texts"] |= set(texts)
            else:
                results["next_page_number"][word] = (
                    1 - results["next_page_number"][word]
                )


def request_rekhta(query, category, page_number=1):
    if category in ["nazm", "ghazal", "couplet"]:
        url = f"https://www.rekhta.org/search/{category}?q={urllib.parse.quote(query)}&lang=ur"
    else:
        url = f"https://www.rekhta.org/CollectionSearchLoading?lang=3&pageType=search%D8%B4%D8%B9%D8%B1content&keyword={urllib.parse.quote(query)}&pageIndex={page_number}&_=1676106622519"

    page = requests.get(url).text

    return page


def extract_rekhta_texts(page: str) -> Dict[str, Set[str]]:
    soup = BeautifulSoup(page, "html.parser")
    cards = soup.find_all("div", {"class": "genricMatchCard"})

    results = {}

    for card in cards:
        category = card.h5.text.strip()
        content = "".join(map(str, card.p.a.contents)).replace("<br/>", "\n")
        if category not in results:
            results[category] = set()
        results[category].add(content)

    return results


def merge_rekhta_results(recieving_dict, sending_dict):
    for k, v in sending_dict.items():
        if k not in recieving_dict:
            recieving_dict[k] = set()
        recieving_dict[k] |= v


def get_rekhta_texts(
    query: str,
    category: str,
    page_no: int = 1,
):
    response = request_rekhta(query, category, page_no)
    texts = extract_rekhta_texts(response)

    return texts


def scraping_function_for_rekhta(
    word: str,
    category: str,
    results: Dict[str, Union[Dict[str, int], Dict[str, Set[str]]]],
):
    if results["next_page_number"][word] > 0:
        texts = get_rekhta_texts(word, category, results["next_page_number"][word])
        with threading.Lock():
            if len(texts) > 0:
                results["next_page_number"][word] += 1
                merge_rekhta_results(results["texts"], texts)
            else:
                results["next_page_number"][word] = (
                    1 - results["next_page_number"][word]
                )


def request_wikipedia(word, limit=10000, lang="ur", namespaces=None):
    if namespaces is None:
        namespaces = "ns0=1&ns1=1&ns10=1&ns100=1&ns101=1&ns102=1&ns103=1&ns11=1&ns118=1&ns119=1&ns12=1&ns13=1&ns14=1&ns15=1&ns2=1&ns2300=1&ns2301=1&ns2302=1&ns2303=1&ns3=1&ns4=1&ns5=1&ns6=1&ns7=1&ns710=1&ns711=1&ns8=1&ns828=1&ns829=1&ns9=1"

    url = f"""https://{lang}.wikipedia.org/w/index.php?limit={limit}&fulltext=1&{namespaces}&search={urllib.parse.quote(f'"{word}"')}"""
    page = requests.get(url).text

    return page


def extract_wikipedia_texts(page: str) -> List[str]:
    soup = BeautifulSoup(page, "html.parser")
    cards = soup.find_all("div", {"class": "searchresult"})

    texts = [card.text.strip() for card in cards]

    return texts


def get_wikipedia_texts(word, limit=10000, lang="ur", namespaces=None) -> List[str]:
    page = request_wikipedia(word, limit=limit, lang=lang, namespaces=namespaces)
    texts = extract_wikipedia_texts(page)

    return texts


def scraping_function_for_wikipedia(
    word: str,
    limit: int,
    lang: str,
    results: Dict[str, Union[Dict[str, int], Set[str]]],
) -> None:
    texts = get_wikipedia_texts(word, limit=limit, lang=lang)
    with threading.Lock():
        if word not in results["n_results"]:
            results["n_results"][word] = 0
        results["n_results"][word] += len(texts)
        results["texts"] |= set(texts)



from sign_language_translator.models import language_models, sign_to_text, text_to_sign
from sign_language_translator.models.text_to_sign import ConcatenativeSynthesis
from sign_language_translator.models.utils import get_model

__all__ = [
    "get_model",
    "language_models",
    "sign_to_text",
    "text_to_sign",
    "ConcatenativeSynthesis",
]



"""defines settings constants for global context
"""

from os.path import join, dirname


class Settings:
    """defines settings constants for global context"""

    DATASET_ROOT_DIRECTORY: str = join(
        dirname(dirname(__file__)), "sign-language-datasets"
    )

    FILENAME_SEPARATOR = "_"
    FILENAME_CONNECTOR = "-"



"""Write new texts with a statistical language model.
"""

import json
import random
from collections import Counter
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Tuple

import numpy

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.text.metrics import normpdf
from sign_language_translator.text.utils import make_ngrams


class SimpleLanguageModel(LanguageModel):
    def __init__(
        self,
        window_size=1,
        unknown_token="<unk>",
        name=None,
        sampling_temperature=1.0,
    ) -> None:
        super().__init__(unknown_token=unknown_token, name=name)
        self.window_size = window_size

        self.NEXT_TOKEN = "NEXT_TOKEN"
        self.WEIGHTS = "WEIGHTS"

        self.sep = "|||"
        self.n_parameters = 0
        self.sampling_temperature = sampling_temperature

    def _to_key_datatype(self, item: Iterable) -> Tuple:
        return tuple(item)

    def _count_ngrams(
        self, training_corpus: List[Iterable], n: int
    ) -> Dict[Tuple, int]:
        all_ngrams = [
            self._to_key_datatype(ngram)
            for sequence in training_corpus
            for ngram in make_ngrams(sequence, n)
        ]
        counts = dict(Counter(all_ngrams))
        return counts

    def _group_by_context(self, counts: Dict[Tuple, int]):
        grouped = {}
        for ngram, freq in counts.items():
            context = ngram[:-1]
            next_token = ngram[-1]

            if context not in grouped:
                grouped[context] = {self.NEXT_TOKEN: [], self.WEIGHTS: []}

            grouped[context][self.NEXT_TOKEN].append(next_token)
            grouped[context][self.WEIGHTS].append(freq)

        return grouped

    def train(self, training_corpus):
        self.fit(training_corpus)

    def fit(self, training_corpus) -> None:
        counts = self._count_ngrams(training_corpus, self.window_size + 1)
        counts = self._group_by_context(counts)

        # normalize frequencies into probabilities
        for context in counts:
            total = sum(counts[context][self.WEIGHTS])
            counts[context][self.WEIGHTS] = [
                freq / total for freq in counts[context][self.WEIGHTS]
            ]

        self.model: Dict[Tuple, Dict[str, List]] = counts
        self.n_parameters = self._count_parameters()
        (
            self.data_distribution_mean,
            self.data_distribution_std,
        ) = self._calculate_data_distribution(training_corpus)

    def finetune(self, training_corpus, weightage: float):
        assert 0.0 <= weightage <= 1.0

        old_model = deepcopy(self.model)
        self.fit(training_corpus)

        for context in old_model:
            if context not in self.model:
                self.model[context] = old_model[context]
            else:
                old_weights = dict(
                    zip(
                        old_model[context][self.NEXT_TOKEN],
                        old_model[context][self.WEIGHTS],
                    )
                )

                for i in range(len(self.model[context][self.NEXT_TOKEN])):
                    next_token = self.model[context][self.NEXT_TOKEN][i]
                    old_w = old_weights.pop(next_token, 0.0)
                    self.model[context][self.WEIGHTS][i] *= weightage
                    self.model[context][self.WEIGHTS][i] += old_w * (1 - weightage)

                for next_token, old_w in old_weights.items():
                    self.model[context][self.NEXT_TOKEN].append(next_token)
                    self.model[context][self.WEIGHTS].append(old_w * (1 - weightage))

        self.n_parameters = self._count_parameters()
        (mean, std) = self._calculate_data_distribution(training_corpus)
        self.data_distribution_mean *= (1-weightage)
        self.data_distribution_mean += weightage * mean
        self.data_distribution_std *= (1-weightage)
        self.data_distribution_std += weightage * std

    def _count_parameters(self):
        return sum(len(v[self.WEIGHTS]) for v in self.model.values())

    def _choose_index(self, weights: List[float]) -> int:
        """Select an item based on the given probability distribution. Returns the index of the selected item sampled from weighted random distribution.

        Args:
            weights (List[float]): the relative weights corresponding to each index.

        Returns:
            int: The index of the chosen item.
        """
        return random.choices(
            range(len(weights)),
            weights=[w / self.sampling_temperature for w in weights],
            k=1,
        )[0]

    def next(self, context: Iterable) -> Tuple[Iterable, float]:
        """Samples the next token from the learnt distribution based on the given context.

        Args:
            context (Iterable): A piece of sequence like the training examples.

        Returns:
            Tuple[Iterable, float]: The next token and its associated probability. Token has the same type as the context iterable.
        """

        next_tokens, probabilities = self.next_all(context)

        index = self._choose_index(probabilities)
        next_token, probability = next_tokens[index], probabilities[index]

        return (next_token, probability)

    def next_all(self, context: Iterable) -> Tuple[List[Iterable], List[float]]:
        context = self._to_key_datatype(context[len(context) - self.window_size :])

        if context not in self.model:
            next_tokens, probabilities = [self.unknown_token], [1.0]
        else:
            next_tokens = self.model[context][self.NEXT_TOKEN]
            probabilities = self.model[context][self.WEIGHTS]

        return next_tokens, probabilities

    def load(self, model_path: str, model_data: str = None) -> None:
        """Deserializes the model (from json).

        Args:
            model_path (str): The source file path.
        """

        if model_data is None:
            with open(model_path, "r") as f:
                model_data = json.load(f)

        self.window_size = model_data["window_size"]
        self.unknown_token = model_data["unknown_token"]

        self.NEXT_TOKEN = model_data["NEXT_TOKEN"]
        self.WEIGHTS = model_data["WEIGHTS"]
        self.sep = model_data["sep"]
        self.sampling_temperature = model_data["sampling_temperature"]

        self.model = {
            self._to_key_datatype(context.split(self.sep)): next_token_and_weights
            for context, next_token_and_weights in model_data["model"].items()
        }

    def save(self, model_path: str, indent=None, ensure_ascii=False) -> None:
        """Serializes the model (as json).

        Args:
            model_path (str): The target file path. It will silently overwrite if a file already exists at this path.
        """

        model_data = dict()
        model_data["window_size"] = self.window_size
        model_data["unknown_token"] = self.unknown_token

        model_data["NEXT_TOKEN"] = self.NEXT_TOKEN
        model_data["WEIGHTS"] = self.WEIGHTS
        model_data["sep"] = self.sep
        model_data["sampling_temperature"] = self.sampling_temperature

        model_data["model"] = {self.sep.join(k): v for k, v in self.model.items()}

        with open(model_path, "w") as f:
            json.dump(model_data, indent=indent, ensure_ascii=ensure_ascii)

    def __str__(self) -> str:
        return f"Simple LM: {super().__str__()}, window={self.window_size}, params={self.n_parameters}"

    def probability(self, token_sequence, return_log_probability=True):
        probabilities = []
        for i in range(self.window_size, len(token_sequence) - 1):
            context = self._to_key_datatype(token_sequence[i - self.window_size : i])
            next_token = token_sequence[i]
            if next_token not in self.model.get(context,{}).get(self.NEXT_TOKEN, []):
                return float("-inf") if return_log_probability else 0.0
            _idx = self.model[context][self.NEXT_TOKEN].index(next_token)
            probabilities.append(self.model[context][self.WEIGHTS][_idx])

        probability = (
            numpy.prod(probabilities)
            if not return_log_probability
            else numpy.log(probabilities).sum()
        )
        return probability

    def _calculate_data_distribution(self, corpus):
        log_probabilities = [
            self.probability(item, return_log_probability=True)
            for item in corpus
        ]
        mean = numpy.mean(log_probabilities)
        std = numpy.std(log_probabilities)
        return mean, std

    def calculate_likelihood(self, token_sequence):
        log_prob = self.probability(
            token_sequence, return_log_probability=True
        )
        likelihood = normpdf(
            log_prob,
            self.data_distribution_mean,
            self.data_distribution_std,
        )
        return likelihood
